"""demixing utility"""
import glob
import os

from aind_ophys_pipeline_utils.pipeline_util import (
    read_json,
    remove_roi_nan,
    get_motion_corrected_movie,
    write_h5,
    read_h5,
    make_output_directory,
)


def create_input_json(input_dir: str, output_dir: str, plane: str) -> dict:
    """Creates input json for demixing

    Args:
        input_dir (str): input directory, data directory (co)
        output_dir (str): output directory, results directory (co)
        plane (str): plane depth

    Returns:
        dict: input json for demixing
    """
    h5_motion_corrected_movie = get_motion_corrected_movie(input_dir)
    try:
        extract_traces = glob.glob(
            f"{input_dir}/Other_[0-9]*/Other/ophys/planes/{plane}/*extract_traces_v2.json"
        )[0]
    except IndexError:
        extract_traces = glob.glob(f"{input_dir}/*extract_traces_v2.json")[0]
    try:
        roi_traces = glob.glob(
            f"{input_dir}/Other_[0-9]*/Other/ophys/planes/{plane}/*roi_traces.h5"
        )[0]
    except IndexError:
        roi_traces = glob.glob(f"{input_dir}/*roi_traces.h5")[0]
    rois = read_h5(roi_traces, "data", "roi_names")
    roi_data = rois["data"]
    roi_names = rois["roi_names"]
    traces = read_json(extract_traces)
    mod_roi_data, _ = remove_roi_nan(traces, roi_names, roi_data)
    os.makedirs("./scratch", exist_ok=True)
    write_h5("./scratch/roi_traces.h5", roi_names=roi_names, data=mod_roi_data)
    output_dir = make_output_directory(output_dir, h5_motion_corrected_movie)
    input_data = dict(
        movie_h5=h5_motion_corrected_movie,
        traces_h5="./scratch/roi_traces.h5",
        output_file=os.path.join(output_dir, f"{plane}um_demixing_output.h5"),
        output_json=os.path.join(output_dir, f"{plane}um_demixing_output.json"),
        roi_masks=traces,
    )

    return input_data



"""Trace Extraction Utility"""
import glob
import os
import json

from aind_ophys_pipeline_utils.pipeline_util import (
    get_motion_corrected_movie,
    add_motion_border,
    clean_roi_table,
    make_output_directory,
)


def create_input_json(input_dir: str, output_dir: str, plane: str) -> dict:
    """Creates input json for extract_traces

    Args:
        input_dir (str): input directory, data directory (co)
        output_dir (str): output directory, results directory (co)
        plane (str): plane depth

    Returns:
        dict: input json for extract_traces
    """
    h5_motion_corrected_movie = get_motion_corrected_movie(input_dir)
    rigid_motion_transform_csv = "*um_suite2p_rigid_motion_transform.csv"
    segmentation_postprocess_json = "*um_suite2p_segmentation_postprocess_rois.json"
    try:
        capsule_directory = f"Other_[0-9]*/Other/ophys/planes/{plane}"
        rigid_motion_transform = glob.glob(
            f"{input_dir}/{capsule_directory}/{rigid_motion_transform_csv}"
        )[0]
    except IndexError:
        rigid_motion_transform = glob.glob(f"{input_dir}/{rigid_motion_transform_csv}")[0]
    try:
        rois = glob.glob(f"{input_dir}/{capsule_directory}/{segmentation_postprocess_json}")[0]
    except IndexError:
        rois = glob.glob(f"{input_dir}/{segmentation_postprocess_json}")[0]
    with open(rois, "r") as f:
        rois = json.load(f)
    output_dir = make_output_directory(output_dir, h5_motion_corrected_movie)
    input_data = dict(
        storage_directory=str(output_dir),
        motion_corrected_stack=h5_motion_corrected_movie,
        output_json=os.path.join(output_dir, f"{plane}um_extract_traces.json"),
        log_0=rigid_motion_transform,
        rois=clean_roi_table(rois),
    )
    input_data = add_motion_border(rigid_motion_transform, input_data)

    return input_data



"""conftest for testing aind-ophys-pipeline-utility"""
from typing import Union
from pathlib import Path
import tempfile

import h5py
import pytest
import json

# import atexit
# import shutil

import numpy as np


class HelperFunctions(object):
    """Helper functions for directoy and file creation"""

    @staticmethod
    def raw_movie_and_platform_dir(child_dir: Union[str, Path]):
        """
        Create a temporary directory to write data to

        """
        temp_dir = tempfile.mkdtemp()
        base_dir = Path(temp_dir)
        base_dir = base_dir / child_dir
        base_dir.mkdir()

        full_dir = base_dir / "Other_614171_2023-03-03_12-00-00"
        full_dir.mkdir()
        full_dir = full_dir / "Other"
        full_dir.mkdir()
        platform_dir = full_dir / "ophys"
        platform_dir.mkdir()
        full_dir = platform_dir / "planes"
        full_dir.mkdir()
        full_dir = full_dir / "300"
        full_dir.mkdir()

        return base_dir, platform_dir, full_dir

    @staticmethod
    def create_input_dir(data_dir: Union[str, Path], child_dir: Union[str, Path]):
        """create statndard directory"""
        temp_dir = tempfile.mkdtemp()
        base_dir = Path(temp_dir)
        base_dir = base_dir / data_dir
        base_dir.mkdir()
        base_dir = base_dir / child_dir
        base_dir.mkdir()
        # def remove_temp_dir():
        #     """Remove the temporary directory"""
        #     shutil.rmtree(temp_dir, ignore_errors=True)

        # atexit.register(remove_temp_dir)
        return base_dir

    @staticmethod
    def create_results_dir(results_dir: Union[str, Path], child_dir: Union[str, Path]):
        """
        Create a temporary directory to write data to

        """
        temp_dir = tempfile.mkdtemp()
        base_dir = Path(temp_dir)
        base_dir = base_dir / results_dir
        base_dir.mkdir()
        base_dir = base_dir / child_dir
        return base_dir


@pytest.fixture
def helper_functions():
    """
    See solution to making helper functions available across
    a pytest module in
    https://stackoverflow.com/questions/33508060/create-and-import-helper-functions-in-tests-without-creating-packages-in-test-di
    """
    return HelperFunctions


@pytest.fixture
def platform_json():
    """platform json data for testing"""
    return {"imaging_plane_groups": [{"acquisition_framerate_Hz": 30.0}]}


@pytest.fixture
def rois():
    """segmented roi masks for testing"""
    return [
        {
            "x": 222,
            "y": 365,
            "width": 12,
            "height": 18,
            "mask_matrix": [
                [False, False, False, False, True, True, True, False, False, False, False, False],
                [False, False, False, True, True, True, True, True, False, False, False, False],
                [False, False, False, True, True, True, True, True, True, False, False, False],
                [False, False, False, True, True, True, True, True, True, True, True, False],
                [False, False, False, True, True, True, True, True, True, True, True, True],
                [False, False, True, True, True, True, True, True, True, True, True, True],
                [False, False, True, True, True, True, True, True, True, True, True, True],
                [False, True, True, True, True, True, True, True, True, True, True, True],
                [False, True, True, True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True, True, True, True],
                [True, True, True, True, True, True, True, True, True, True, True, True],
                [False, True, True, True, True, True, True, True, True, True, True, True],
                [False, True, True, True, True, True, True, True, True, True, True, True],
                [False, False, True, True, True, True, True, True, True, True, True, True],
                [False, False, False, True, True, True, True, True, True, True, True, False],
                [False, False, False, False, True, True, True, True, True, True, True, False],
                [False, False, False, False, False, True, True, True, True, True, False, False],
                [False, False, False, False, False, False, True, True, True, False, False, False],
            ],
            "valid_roi": True,
            "mask_image_plane": 0,
            "exclusion_labels": [],
            "id": 0,
            "max_correction_up": 0,
            "max_correction_down": 0,
            "max_correction_left": 0,
            "max_correction_right": 0,
        },
        {
            "x": 70,
            "y": 273,
            "width": 14,
            "height": 10,
            "mask_matrix": [
                [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                ],
                [
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                ],
                [
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
                [
                    False,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
                [
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                ],
                [
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                ],
                [
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    False,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                ],
                [
                    False,
                    False,
                    False,
                    False,
                    False,
                    True,
                    True,
                    True,
                    False,
                    False,
                    False,
                    False,
                    False,
                    False,
                ],
            ],
            "valid_roi": True,
            "mask_image_plane": 0,
            "exclusion_labels": [],
            "id": 1,
            "max_correction_up": 0,
            "max_correction_down": 0,
            "max_correction_left": 0,
            "max_correction_right": 0,
        },
    ]


@pytest.fixture
def roi_trace():
    """roi traces for testing"""
    return {
        "data": [
            np.array([22, 33, 44, 11, 39]).astype(np.float64),
            np.array([22, 33, 44, 11, 39]).astype(np.float64),
        ],
        "roi_names": ["0", "1"],
    }


@pytest.fixture
def extract_traces():
    """extract traces for testing"""
    return {
        "exclusion_labels": [
            {"exclusion_label_name": "motion_border", "roi_id": 0},
            {"exclusion_label_name": "motion_border", "roi_id": 1},
        ]
    }


@pytest.fixture
def rigid_motion_transform():
    """rigid motion transform for testing"""
    return [
        "framenumber,x,y,x_pre_clip,y_preclip,correlation,is_valid\n",
        "0,5,14,5,14,0.01,True\n",
        "1,5,13,5,13,0.012572901,True\n",
    ]


@pytest.fixture
def ophys_files_capsule_pre_process(platform_json):
    """create fileset for capsule testing"""
    base_dir, platform_dir, full_dir = HelperFunctions.raw_movie_and_platform_dir("data/")
    raw_movie_fp = full_dir / "300um.h5"
    with h5py.File(raw_movie_fp, "w") as f:
        f.create_dataset("data", data=np.random.rand(10, 20, 20))
    platform_json_fp = platform_dir / "platform.json"
    with open(platform_json_fp, "w") as f:
        json.dump(platform_json, f)
    return base_dir, full_dir


@pytest.fixture
def ophys_files_capsule_post_process(
    platform_json,
    rois,
    roi_trace,
    extract_traces,
    rigid_motion_transform
):
    """create fileset for capsule testing"""
    post_process_dir = HelperFunctions.create_input_dir(
        "data/", "Other_614171_2023-03-03_12-00-00/"
    )
    platform_json_fp = post_process_dir / "platform.json"
    with open(platform_json_fp, "w") as f:
        json.dump(platform_json, f)
    fn_motion_corrected = post_process_dir / "300um_motion_output.h5"
    with h5py.File(fn_motion_corrected, "w") as f:
        f.create_dataset("data", data=np.random.rand(10, 20, 20))
    rois_fp = post_process_dir / "300um_suite2p_segmentation_postprocess_rois.json"
    print(rois_fp)
    with open(rois_fp, "w") as f:
        json.dump(rois, f)
    rigid_motion_transforms_fp = post_process_dir / "300um_suite2p_rigid_motion_transform.csv"
    with open(rigid_motion_transforms_fp, "w") as f:
        for i in rigid_motion_transform:
            f.write(i)
    roi_traces = post_process_dir / "300um_roi_traces.h5"
    with h5py.File(roi_traces, "w") as f:
        for k, v in roi_trace.items():
            f.create_dataset(k, data=v)
    extract_traces_fp = post_process_dir / "300um_extract_traces_v2.json"
    with open(extract_traces_fp, "w") as f:
        json.dump(extract_traces, f)
    neuropil_trace_h5 = post_process_dir / "300um_neuropil_trace.h5"
    with h5py.File(neuropil_trace_h5, "w") as f:
        f.create_dataset("data", data=np.random.rand(10, 20, 20))
    neuropil_trace_json = post_process_dir / "300um_neuropil_trace_output.json"
    with open(neuropil_trace_json, "w") as f:
        json.dump(platform_json, f)
    return post_process_dir



"""pipeline utility functions"""
import h5py as h5
import json
from datetime import datetime as dt
import pytz
import os
import glob
import subprocess
import numpy as np
from collections import namedtuple
import pandas as pd
from pathlib import Path
import platform
import re

from aind_ophys_utils import motion_border_utils


def write_h5(fp: str, **kwargs: dict) -> bool:
    """Writes data to h5 file

    Args:
        fp (str): filepath to write data to
        kwargs (dict): data to write to h5 file

    Raises:
        Exception: Error writing h5 file
    Returns:
        bool: True if successful
    """
    try:
        with h5.File(fp, "w") as h:
            for k, v in kwargs.items():
                h.create_dataset(k, data=v)
        return True
    except Exception as e:
        raise Exception(f"Error writing h5 file: {e}")


def read_h5(fp: str, *args: list) -> dict:
    """Reads a list of keys from an h5 file

    Args:
        fp (str): filepath to read data from

    Returns:
        dict: data from h5 file in dictionary format
    """
    data = {}
    h5data = h5.File(fp, "r")
    for i in args:
        data[i] = h5data[i][()]
    h5data.close()
    return data


def write_json(fp: str, data: dict) -> bool:
    """Write data to json file
    Args:
        fp (str): filepath to write data to
        data (dict): data to write to json file

    Raises:
        Exception: Error writing json file
    Returns:
        bool: True if successful
    """
    try:
        with open(fp, "w") as j:
            json.dump(data, j, indent=2)
        return True
    except Exception as e:
        raise Exception(f"Error writing json file: {e}")


def read_json(fp: str) -> dict:
    """Read json file

    Args:
        fp (str): filepath to read data from

    Raises:
        Exception: Error reading json file
    Returns:
        dict: data from json file in dictionary format
    """
    try:
        with open(fp, "r") as j:
            return json.load(j)
    except Exception as e:
        raise Exception(f"Error reading json file: {e}")


def now() -> str:
    """Generates string with current date and time in PST

    Returns:
        str: YYYY-MM-DD_HH-MM-SS
    """
    current_dt = dt.now(tz=pytz.timezone("America/Los_Angeles"))
    return f"{current_dt.strftime('Y-m-d')}_{current_dt.strftime('H:M:S')}"


def create_results_folder(output_file_path: str, input_file_path: str) -> str:
    """Creates a results folder to store processed data

    Args:
        output_file_path (str): path to output folder
        input_file_path (str): path to input file

    Raises:
        OSError: Error creating results folder
    Returns:
        str: path to results folder
    """
    try:
        output_file_path = f"{output_file_path}_{get_file_name(input_file_path)}_processed_{now}"
        os.makedirs(output_file_path, exist_ok=True)
        return output_file_path
    except OSError as error:
        raise OSError(f"Error: {error}")


def get_basename(filename: str) -> str:
    """gets basename from filename for results data naming process

    Args:
        filename (str): filename to get basename from

    Raises:
        IndexError: could not get basename
    Returns:
        str: basename
    """
    try:
        return filename.split("/")[-1].split(".")[0]
    except IndexError:
        raise IndexError(f"Error: {filename} is not a valid file path")


def get_file_name(file_path: str) -> str:
    """gets basename of file path

    Args:
        file_path (str): file path to get file name from

    Raises:
        IndexError: could not get file name
    Returns:
        str: file name
    """
    try:
        return file_path.split("/")[-1].split(".")[0]
    except IndexError:
        raise IndexError(f"Error: {file_path} is not a valid file path")


def get_plane(input_dir: Path) -> str:
    """Get the imaging plane from the session directory to know how to save the data

    Args:
        input_dir (str): directory where file is located

    Raises:
        Exception: if no plane depth is found
    Returns:
        plane (str): plane depth or None if no plane depth is found
    """

    try:
        filepath = glob.glob(f"{input_dir}/Other_[0-9]*/Other/ophys/planes/*")[0]
        if platform.system() == "Windows":
            plane = filepath.split("\\")[-1]
        else:
            plane = filepath.split("/")[-1]
    except IndexError:
        try:
            plane = re.findall(r"\d+um", glob.glob(f"{input_dir}/*")[0])[0].split("um")[0]
        except IndexError:
            raise IndexError("No plane depth found")
    assert int(plane)
    return plane


def get_motion_corrected_movie(input_dir: str) -> str:
    """Retrieves the motion corrected, single plane movie from the /data directory

    Args:
        input_dir (str): directory where file is located

    Raises:
        Exception: if no motion corrected movie is found
    Returns:
        h5_movie (str): motion corrected movie path
    """
    try:
        h5_movie = glob.glob(f"{input_dir}/*motion_output.h5")[0]
        return h5_movie
    except IndexError as exc:
        raise Exception(f"Error: {exc}")


def copy_data_to_results(src: str, dest: str) -> None:
    """Copies data from source to destination

    Args:
        src (str): source file path
        dest (str): destination file path

    Raises:
        subprocess.CalledProcessError: Error syncing the data to results folder
    """
    try:
        subprocess.run(["rsync", "-avz", src, dest])
    except subprocess.CalledProcessError as exc:
        raise subprocess.CalledProcessError(f"Error: {exc}")


def clean_roi_table(data: dict, remove_exclusion_labels=False) -> dict:
    """
    Removes unnecessary keys from the roi table

    Args:
        data (dict): roi table data
        remove_exclusion_labels (bool): remove exclusion labels from roi table data

    Returns:
        data (dict): cleaned roi table data
    """
    print(len(data))
    for roi in range(len(data)):
        data[roi].pop("max_correction_down", None)
        data[roi].pop("max_correction_up", None)
        data[roi].pop("max_correction_right", None)
        data[roi].pop("max_correction_left", None)
        if remove_exclusion_labels:
            data[roi].pop("exclusion_labels", None)
        data[roi].pop("mask_image_plane", None)
        data[roi]["mask"] = data[roi].get("mask_matrix", [])
    return data


def add_motion_border(input_csv, data):
    """Finds and adds max correction values assuming a max shift of 30.0 (default value in suite2p)

    Args:
        input_csv (str): rigid transform csv file output from suite2p motion correction
        data (dict): data for the input json file

    Raises:
        Exception: max shift could not be retrieved from csv file
    Returns:
        data (): input json data with max shift values added
    """
    csv_data = pd.read_csv(input_csv)
    max_shift = motion_border_utils.get_max_correction_from_df(csv_data)

    data["motion_border"] = {}

    # add max shift (left, right, up, down) to input json
    data["motion_border"]["x0"] = max_shift.left
    data["motion_border"]["x1"] = max_shift.right
    data["motion_border"]["y1"] = max_shift.up
    data["motion_border"]["y0"] = max_shift.down
    return data


def get_frame_rate_platform_json(input_dir: str) -> float:
    """Get the frame rate from the platform json file.
    Platform json will need to get copied to each data directory throughout the pipeline

    Args:
        input_dir (str): directory where file is located

    Raises:
        Exception: if no platform json file is found
    Returns:
        frame_rate (float): frame rate
    """
    try:
        try:
            platform_json = glob.glob(f"{input_dir}/Other_[0-9]*/Other/ophys/*platform.json")[0]
        except IndexError:
            platform_json = glob.glob(f"{input_dir}/*platform.json")[0]
        with open(platform_json) as f:
            data = json.load(f)
        frame_rate = data["imaging_plane_groups"][0]["acquisition_framerate_Hz"]
        return frame_rate
    except IndexError as exc:
        raise Exception(f"Error: {exc}")


def remove_roi_nan(trace_exclusions: dict, roi_names: np.array, roi_data: np.array):
    """Removes ROIs that are NaNs from the ROI table

    Args:
        trace_exclusions (dict): trace exclusions file
        roi_names (np.array): roi names
        roi_data (dict): roi data

    Returns:
        roi_data (dict): roi data with NaNs removed
        roi_names (np.array): roi names with NaNs removed
    Raises:
        Exception: if roi_data is empty"""
    try:
        for i in range(len(trace_exclusions["exclusion_labels"])):
            # value = trace_exclusions['exclusion_labels'][i]['roi_id']
            # index = np.where(roi_names == value.encode())[0][0]
            find_nans = np.isnan(roi_data)
            roi_data[find_nans] = 0
            # roi_data = np.delete(roi_data, index, 0)
            # roi_names = np.delete(roi_names, index, 0)
    except Exception as exc:
        raise Exception(f"Error: {exc}")
    return roi_data, roi_names


def make_output_directory(output_dir: str, h5_file: str) -> str:
    """Creates the output directory if it does not exist

    Args:
        output_dir (str): output directory
        h5_file (str): h5 file path

    Returns:
        output_dir (str): output directory
    """
    exp_to_match = r"Other_\d{6}_\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}"
    parent_dir = re.findall(exp_to_match, h5_file)[0]
    output_dir = os.path.join(output_dir, parent_dir)
    os.makedirs(output_dir, exist_ok=True)
    return output_dir


# see comment in motion_border_from_max_frame_shift for an explanation
# of the difference between these two structures

MaxFrameShift = namedtuple("MaxFrameShift", ["left", "right", "up", "down"])

MotionBorder = namedtuple("MotionBorder", ["top", "bottom", "left_side", "right_side"])


def get_max_correction_values(
    x_series: pd.Series, y_series: pd.Series, max_shift: float = 30.0
) -> MaxFrameShift:
    """
    Gets the max correction values in the cardinal directions from a series
    of correction values in the x and y directions
    Parameters
    ----------
    x_series: pd.Series:
        A series of movements in the x direction
    y_series: pd.Series:
        A series of movements in the y direction
    max_shift: float
        Maximum shift to allow when considering motion correction. Any
        larger shifts are considered outliers (only absolute value matters).

    For deprecated implementation see:
    allensdk.internal.brain_observatory.roi_filter_utils.calculate_max_border

    Returns
    -------
    MaxFrameShift
        A named tuple containing the maximum correction values found during
        motion correction workflow step. Saved with the following direction
        order [left, right, up, down].

    """
    # take abs of max shift as we are considering both positive and negative
    # directions
    max_shift = abs(max_shift)

    # filter based out analomies based on maximum_shift
    x_no_outliers = x_series[(x_series >= -max_shift) & (x_series <= max_shift)]
    y_no_outliers = y_series[(y_series >= -max_shift) & (y_series <= max_shift)]
    # calculate max border shifts
    right_shift = -1 * x_no_outliers.min()
    left_shift = x_no_outliers.max()
    down_shift = -1 * y_no_outliers.min()
    up_shift = y_no_outliers.max()

    max_shift = MaxFrameShift(left=left_shift, right=right_shift, up=up_shift, down=down_shift)

    # check if all exist
    if np.any(np.isnan(np.array(max_shift))):
        raise ValueError(
            "One or more motion correction shifts "
            "was found to be Nan, max shift found: "
            f"{max_shift}, with max_shift {max_shift}"
        )

    return max_shift


def get_max_correction_from_file(input_csv: Path, max_shift: float = 30.0) -> MaxFrameShift:
    """

    Parameters
    ----------
    input_csv: Path
        Path to motion correction values for each frame stored in .csv format.
        This .csv file is expected to have a header row of either:
        ['framenumber','x','y','correlation','kalman_x', 'kalman_y'] or
        ['framenumber','x','y','correlation','input_x','input_y','kalman_x',
         'kalman_y','algorithm','type']
    max_shift: float
        Maximum shift to allow when considering motion correction. Any
        larger shifts are considered outliers.

    Returns
    -------
    max_shift
        A named tuple containing the maximum correction values found during
        motion correction workflow step. Saved with the following direction
        order [left, right, up, down].

    """
    motion_correction_df = pd.read_csv(input_csv)
    max_shift = get_max_correction_values(
        x_series=motion_correction_df["x"].astype("float"),
        y_series=motion_correction_df["y"].astype("float"),
        max_shift=max_shift,
    )
    return max_shift


def motion_border_from_max_shift(max_shift: MaxFrameShift) -> MotionBorder:
    """
    Find the MotionBorder that corresponds to a given
    MaxFrameShift
    """

    # The difference between MaxFrameShift and MotionBorder is that
    # MaxFrame shift records the maximum shift that was applied to
    # a movie in a given direction during motion correction. This could,
    # in principle, be negative (if a movie was only ever shifted up, it's
    # maximum down shift will be negative). MotionBorder is the positive
    # (or zero) number of pixels to ignore at the edge of a field of view.
    # In addition to the fact that MotionBorder can only be positive, there
    # is a transposition. If a movie is only ever shifted up, there should
    # be a non-zero motion border at the bottom, since those pixels were
    # either wrapped or padded by motion correction, but there should be
    # no motion border at the top (assuming you trust your motion correction
    # algorithm).

    result = MotionBorder(
        bottom=max(max_shift.up, 0),
        top=max(max_shift.down, 0),
        left_side=max(max_shift.right, 0),
        right_side=max(max_shift.left, 0),
    )

    return result



"""Pars command line args"""
import argparse

from aind_ophys_pipeline_utils.pipeline_util import get_plane, write_json, copy_data_to_results
from aind_ophys_pipeline_utils import (
    demixing_utils,
    dff_utils,
    neuropil_correction_utils,
    segmentation_utils,
    suite2p_motion_correction_utils,
    trace_extraction_utils,
)

parser = argparse.ArgumentParser()
parser.add_argument("-i", "--input_dir", type=str, help="Input directory")
parser.add_argument("-o", "--output_dir", type=str, help="Output directory")
parser.add_argument("-p", "--plane", type=int, help="Plane depth", default=None)
parser.add_argument(
    "-a", "--action", type=str, help="Action to perform", default="create-input-json"
)
parser.add_argument("-c", "--capsule", type=str, help="Capsule to be run")


args = parser.parse_args()
input_dir = args.input_dir
output_dir = args.output_dir
capsule = args.capsule
action = args.action

if not args.plane:
    plane = get_plane(input_dir)
else:
    plane = args.plane

capsules = dict(
    demixing=demixing_utils,
    dff=dff_utils,
    neuropil_correction=neuropil_correction_utils,
    segmentation=segmentation_utils,
    suite2p_motion_correction=suite2p_motion_correction_utils,
    trace_extraction=trace_extraction_utils,
)
if action == "create-input-json":
    data = capsules[capsule].create_input_json(input_dir, output_dir, plane)
    write_json(input_dir + "input.json", data)
elif action == "copy-data-to-results":
    copy_data_to_results(input_dir, output_dir)



"""suite2p motion utility"""
import glob
import os
from pathlib import Path

from aind_ophys_pipeline_utils.pipeline_util import (
    make_output_directory,
    get_frame_rate_platform_json,
)


def get_base_dir(input_dir: Path) -> str:
    """Returns base directory from input directory specific to this action
    (i.e. suite2p motion correction)

    Args:
        input_dir (Path): path to data directory

    Returns:
        str: base directory
    """
    return str(input_dir.parent.parent.parent)


def create_input_json(input: str, output_dir: str, plane: str) -> dict:
    """Generates suite2p motion correction input json for a given plane

    Args:
        input (str): path to data directory or path to raw movie
        output_dir (str): path to results directory
        plane (int, optional): Plane to process if session directory is passed. Defaults to None.

    Returns:
        dict: suite2p motion correction input json
    """
    try:
        h5_file = glob.glob(f"{input}/Other_[0-9]*/Other/ophys/planes/{plane}/*um.h5")[0]
        base_dir = input
    except IndexError:
        h5_file = input
        base_dir = get_base_dir(Path(input))
    output_dir = make_output_directory(output_dir, h5_file)

    frame_rate_hz = get_frame_rate_platform_json(base_dir)
    return {
        "h5py": h5_file,
        "movie_frame_rate_hz": frame_rate_hz,
        "motion_corrected_output": os.path.join(output_dir, f"{plane}um_suite2p_motion_output.h5"),
        "motion_diagnostics_output": os.path.join(
            output_dir, f"{plane}um_suite2p_rigid_motion_transform.csv"
        ),
        "max_projection_output": os.path.join(
            output_dir, f"{plane}um_suite2p_maximum_projection.png"
        ),
        "avg_projection_output": os.path.join(
            output_dir, f"{plane}um_suite2p_average_projection.png"
        ),
        "registration_summary_output": os.path.join(
            output_dir, f"{plane}um_suite2p_registration_summary.png"
        ),
        "motion_correction_preview_output": os.path.join(
            output_dir, f"{plane}um_suite2p_motion_preview.webm"
        ),
        "output_json": os.path.join(output_dir, f"{plane}um_suite2p_motion_correction_output.json"),
    }



"""Tests neuropil_correction_utils"""
import pytest

from aind_ophys_pipeline_utils import neuropil_correction_utils
from aind_ophys_pipeline_utils.pipeline_util import get_plane


@pytest.mark.parametrize(
    "neuropil_trace_h5, roi_trace_h5, neuropil_trace_output_json",
    [
        ("300um_neuropil_trace.h5", "300um_roi_traces.h5", "300um_neuropil_trace_output.json"),
    ],
)
def test_create_input_json(
    neuropil_trace_h5,
    roi_trace_h5,
    neuropil_trace_output_json,
    ophys_files_capsule_post_process,
    helper_functions,
):
    """Test neuropil correction input json"""
    postprocess_dir = ophys_files_capsule_post_process
    plane = get_plane(postprocess_dir)
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    expected = {
        "neuropil_trace_file": str(postprocess_dir / neuropil_trace_h5),
        "roi_trace_file": str(postprocess_dir / roi_trace_h5),
        "storage_directory": str(output_dir),
        "output_json": str(output_dir / neuropil_trace_output_json),
    }

    actual = neuropil_correction_utils.create_input_json(postprocess_dir, output_dir.parent, plane)
    assert actual == expected



"""Test suite2p motion correction"""
from aind_ophys_pipeline_utils import trace_extraction_utils
from aind_ophys_pipeline_utils.pipeline_util import get_plane, clean_roi_table, add_motion_border
import pytest


@pytest.mark.parametrize(
    "motion_corrected_output",
    [
        ("300um_motion_output.h5"),
    ],
)
def test_create_input_json_capsule(
    motion_corrected_output, ophys_files_capsule_post_process, helper_functions, rois
):
    """Test trace extraction input json"""

    post_process_dir = ophys_files_capsule_post_process
    plane = get_plane(post_process_dir)
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    motion_corrected_dir = post_process_dir
    input_data = {
        "storage_directory": str(output_dir),
        "motion_corrected_stack": str(motion_corrected_dir / motion_corrected_output),
        "output_json": str(output_dir / f"{plane}um_extract_traces.json"),
        "log_0": str(motion_corrected_dir / "300um_suite2p_rigid_motion_transform.csv"),
        "rois": clean_roi_table(rois),
    }
    expected = add_motion_border(
        str(motion_corrected_dir / "300um_suite2p_rigid_motion_transform.csv"), input_data
    )
    actual = trace_extraction_utils.create_input_json(post_process_dir, output_dir.parent, plane)
    assert actual == expected



"""Test suite2p motion correction"""
from aind_ophys_pipeline_utils import segmentation_utils
from aind_ophys_pipeline_utils.pipeline_util import get_plane
import pytest


@pytest.mark.parametrize(
    "movie_frame_rate_hz, motion_corrected_output, segmentation_output",
    [
        (30.0, "300um_motion_output.h5", "300um_segmentation_output.json"),
    ],
)
def test_create_input_json_capsule(
    movie_frame_rate_hz,
    motion_corrected_output,
    segmentation_output,
    helper_functions,
    ophys_files_capsule_post_process,
):
    """Test segmentation input json"""
    post_process_dir = ophys_files_capsule_post_process
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    plane = get_plane(post_process_dir)
    expected = {
        "suite2p_args": {
            "h5py": str(post_process_dir / motion_corrected_output),
            "tmp_dir": "/scratch",
            "movie_frame_rate_hz": movie_frame_rate_hz,
        },
        "postprocess_args": {
            "motion_corrected_video": str(post_process_dir / motion_corrected_output),
            "output_json": str(output_dir / segmentation_output),
        },
    }
    actual = segmentation_utils.create_input_json(post_process_dir, output_dir.parent, plane)
    assert actual == expected



"""Tests dff_utils"""
import pytest

from aind_ophys_pipeline_utils import dff_utils
from aind_ophys_pipeline_utils.pipeline_util import get_plane


@pytest.mark.parametrize(
    "neuropil_trace_h5, dff_h5, movie_frame_rate_hz",
    [
        ("300um_neuropil_trace.h5", "300um_dff.h5", 30.0),
    ],
)
def test_create_input_json_capsule(
    neuropil_trace_h5,
    dff_h5,
    movie_frame_rate_hz,
    ophys_files_capsule_post_process,
    helper_functions,
):
    """Test dff input json"""
    postprocess_dir = ophys_files_capsule_post_process
    plane = get_plane(postprocess_dir)
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    expected = {
        "input_file": str(postprocess_dir / neuropil_trace_h5),
        "movie_frame_rate_hz": movie_frame_rate_hz,
        "output_file": str(output_dir / dff_h5),
    }

    actual = dff_utils.create_input_json(postprocess_dir, output_dir.parent, plane)
    assert actual == expected



"""dF/F utility"""
import glob
import os

from aind_ophys_pipeline_utils.pipeline_util import (
    make_output_directory,
    get_frame_rate_platform_json,
)


def create_input_json(input_dir: str, output_dir: str, plane: str) -> dict:
    """Creates input json for neuropil trace processing

    Args:
        input_dir (str): input directory, data directory (co)
        output_dir (str): output directory, results directory (co)
        plane (str): plane depth

    Returns:
        dict: input json for neuropil trace processing
    """
    try:
        neuropil_trace = glob.glob(
            f"{input_dir}/Other_[0-9]*/Other/ophys/planes/{plane}/*neuropil_trace.h5"
        )[0]
    except IndexError:
        neuropil_trace = glob.glob(f"{input_dir}/*neuropil_trace.h5")[0]
    frame_rate_hz = get_frame_rate_platform_json(input_dir)
    output_dir = make_output_directory(output_dir, neuropil_trace)
    input_data = dict(
        input_file=neuropil_trace,
        movie_frame_rate_hz=frame_rate_hz,
        output_file=os.path.join(output_dir, f"{plane}um_dff.h5"),
    )

    return input_data



"""Tests demixing_utils"""
import pytest
import os

from aind_ophys_pipeline_utils import demixing_utils
from aind_ophys_pipeline_utils.pipeline_util import (
    get_plane,
    read_h5,
    remove_roi_nan,
    write_h5,
    read_json,
)


@pytest.mark.parametrize(
    "motion_corrected_output, roi_trace_h5, extract_traces_v2_json",
    [("300um_motion_output.h5", "300um_roi_traces.h5", "300um_extract_traces_v2.json")],
)
def test_create_input_json_capsule(
    motion_corrected_output,
    roi_trace_h5,
    extract_traces_v2_json,
    ophys_files_capsule_post_process,
    helper_functions,
):
    """Test demixing input json"""
    postprocess_dir = ophys_files_capsule_post_process
    motion_corrected_fn = postprocess_dir / motion_corrected_output
    plane = get_plane(postprocess_dir)
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    os.makedirs("./scratch", exist_ok=True)
    roi_traces = postprocess_dir / roi_trace_h5
    extract_traces = postprocess_dir / extract_traces_v2_json
    rois = read_h5(roi_traces, "data", "roi_names")
    roi_data = rois["data"]
    roi_names = rois["roi_names"]
    traces = read_json(extract_traces)
    mod_roi_data, _ = remove_roi_nan(traces, roi_names, roi_data)
    write_h5("./scratch/roi_traces.h5", roi_names=roi_names, data=mod_roi_data)
    expected = {
        "movie_h5": str(motion_corrected_fn),
        "traces_h5": "./scratch/roi_traces.h5",
        "output_file": str(output_dir / f"{plane}um_demixing_output.h5"),
        "output_json": str(output_dir / f"{plane}um_demixing_output.json"),
        "roi_masks": traces,
    }

    actual = demixing_utils.create_input_json(postprocess_dir, output_dir.parent, plane)
    assert actual == expected



"""Test suite2p motion correction"""
from aind_ophys_pipeline_utils import suite2p_motion_correction_utils
from aind_ophys_pipeline_utils.pipeline_util import get_plane
import pytest


@pytest.mark.parametrize(
    "movie_frame_rate_hz, raw_movie, motion_corrected_output, \
        motion_diagnostics_output, max_projection_output, \
        avg_projection_output, registration_summary_output, \
        motion_correction_preview_output, output_json",
    [
        (
            30.0,
            "300um.h5",
            "300um_suite2p_motion_output.h5",
            "300um_suite2p_rigid_motion_transform.csv",
            "300um_suite2p_maximum_projection.png",
            "300um_suite2p_average_projection.png",
            "300um_suite2p_registration_summary.png",
            "300um_suite2p_motion_preview.webm",
            "300um_suite2p_motion_correction_output.json",
        ),
    ],
)
def test_create_input_json(
    movie_frame_rate_hz,
    raw_movie,
    motion_corrected_output,
    motion_diagnostics_output,
    max_projection_output,
    avg_projection_output,
    registration_summary_output,
    motion_correction_preview_output,
    output_json,
    ophys_files_capsule_pre_process,
    helper_functions,
):
    """Test suite2p motion correction input json"""
    base_dir, full_dir = ophys_files_capsule_pre_process
    output_dir = helper_functions.create_results_dir(
        "results/", "Other_614171_2023-03-03_12-00-00/"
    )
    plane = get_plane(full_dir)
    raw_movie_fp = str(full_dir / raw_movie)
    expected = {
        "h5py": raw_movie_fp,
        "movie_frame_rate_hz": movie_frame_rate_hz,
        "motion_corrected_output": str(output_dir / motion_corrected_output),
        "motion_diagnostics_output": str(output_dir / motion_diagnostics_output),
        "max_projection_output": str(output_dir / max_projection_output),
        "avg_projection_output": str(output_dir / avg_projection_output),
        "registration_summary_output": str(output_dir / registration_summary_output),
        "motion_correction_preview_output": str(output_dir / motion_correction_preview_output),
        "output_json": str(output_dir / output_json),
    }
    # Tests capsule input with only base_dir
    actual_capsule = suite2p_motion_correction_utils.create_input_json(
        base_dir, output_dir.parent, plane
    )
    assert actual_capsule == expected
    # Tests pipeline input with full_dir
    actual_pipeline = suite2p_motion_correction_utils.create_input_json(
        raw_movie_fp, output_dir.parent, plane
    )
    assert actual_pipeline == expected



"""Segmentation Utility"""
import os
from aind_ophys_pipeline_utils.pipeline_util import (
    get_motion_corrected_movie,
    make_output_directory,
    get_frame_rate_platform_json,
)


def create_input_json(input_dir: str, output_dir: str, plane: str) -> dict:
    """Generates input json for segment_postprocess module in ophys_etl

    Args:
        input_dir (str): path to data directory
        output_dir (str): path to results directory
        plane (str): plane depth

    Returns:
        dict: suite2p motion correction input json
    """
    print(f"INPUT DIRECTORY {input_dir}")
    motion_corrected_fn = get_motion_corrected_movie(input_dir)
    output_dir = make_output_directory(output_dir, motion_corrected_fn)
    return {
        "suite2p_args": {
            "h5py": motion_corrected_fn,
            "movie_frame_rate_hz": get_frame_rate_platform_json(input_dir),
            "tmp_dir": "/scratch",
        },
        "postprocess_args": {
            "motion_corrected_video": motion_corrected_fn,
            "output_json": os.path.join(output_dir, f"{plane}um_segmentation_output.json"),
        },
    }



"""Neuropil correction utility"""
import glob
import os

from aind_ophys_pipeline_utils.pipeline_util import (
    make_output_directory,
)


def create_input_json(input_dir: str, output_dir: str, plane: str) -> dict:
    """Creates input json for neuropil trace processing

    Args:
        input_dir (str): input directory, data directory (co)
        output_dir (str): output directory, results directory (co)
        plane (str): plane depth

    Returns:
        dict: input json for neuropil trace processing
    """
    try:
        neuropil_trace = glob.glob(
            f"{input_dir}/Other_[0-9]*/Other/ophys/planes/{plane}/*neuropil_trace.h5"
        )[0]
    except IndexError:
        neuropil_trace = glob.glob(f"{input_dir}/*neuropil_trace.h5")[0]
    try:
        roi_traces = glob.glob(
            f"{input_dir}/Other_[0-9]*/Other/ophys/planes/{plane}/*roi_traces.h5"
        )[0]
    except IndexError:
        roi_traces = glob.glob(f"{input_dir}/*roi_traces.h5")[0]
    output_dir = make_output_directory(output_dir, neuropil_trace)
    input_data = dict(
        neuropil_trace_file=neuropil_trace,
        roi_trace_file=roi_traces,
        storage_directory=output_dir,
        output_json=os.path.join(output_dir, f"{plane}um_neuropil_trace_output.json"),
    )

    return input_data
